#####################################################################################################
#
# statex.py -- VCS commit message action and artifact extractor Class hierarchy/mechanism
#             and test script to demonstrate how to extract segments of a commit message
#             that contain artifact ident(s) and action(s)
#
#    Assume that any artifact ident and action verbiage must occur before any other verbiage
#    and that we will _not_ operate on any such likely patterns if they are interrupted by
#    non artifact ident / action text.
#    ie., DE1235 Fixed and wovnerio never tested this  <-- qualifies
#    Completed TA19 TA12 Fixed DE321 deprecated old rounding operation <-- qualifies
#    updated US123 feature by delegating to proxy as demo in S344 <-- qualifies for artifacts only
#    the rogoshus project won't like that I Rejected DE123  <-- qualifies for artifacts only
#    unable to figure out if DS3423 is ever validated <--- does not qualify
#    A message that only mentions artifacts and not any valid action also qualifies
#     it just won't result in the update of the artifact State/ScheduleState attribute
#
#    Collapse all whitespace (space, tab, newline) sequences into a single space,
#    as well as substituting a space for common punctionation chars,
#    then split the message on whitespace, and reverse the order to obtain a stack of words.
#    pop words off the stack until a word
#       doesn't match the ACTION_WORD_PATTERN or ARTIFACT_IDENT_PATTERN
#    have a dict of {action : [artifacts]} as the result
#    if you get the artifacts first, then you have to hold on to them until you get the action
#
#####################################################################################################

import sys
import re

#####################################################################################################

ACTION_WORD_PATTERN    = re.compile(r'[A-Z][a-z]+[a-zA-Z_-]+')
ARTIFACT_IDENT_PATTERN = re.compile(r'(?P<art_prefix>[A-Z]{1,2})(?P<art_num>\d+)')

#####################################################################################################

class ActionsAndArtifactsExtractor(object):
    def __init__(self, artifacts_metadata):
        """
            artifact_specs is a list of dicts with each dict having the following info:
                abbrev :  S | US | DS | DE | TA | TC etc..
                update_field:  State | ScheduleState | CustomField etc...
                valid_values:  Submitted, Open, Defined, Fixed, Closed, Replicated, Rejected, etc.
        """
        self.valid_artifact_abbrev = [key[0] for key in artifacts_metadata.keys()]
        all_actions = [action for sublist in artifacts_metadata.values() for action in sublist]
        self.artifacts_metadata = artifacts_metadata
        self.valid_action_words = list(set(all_actions))
        self.valid_artifact_actions = {key[0]:value for key, value in self.artifacts_metadata.items()}

    def extractActionGroups(self, message):
        """
            Your subclass will need to implement this method to return a dict
            with actions and artifact targets extracted from the message.
        """
        raise NotImplementedError

    service = extractActionGroups

#####################################################################################################

class BasicActionsAndArtifactsExtractor(ActionsAndArtifactsExtractor):
    def extractActionGroups(self, message):
        """
            Given a string (possibly multiline...) with a commit message that may contain
            references to Rally artifacts and intended state related update values,
            extract those references and values.
            Return a dict keyed by update value with the list of Rally artifacts
            to be updated with that update value for each update value.
            Example:
               {'Completed' : ['TA12', 'TA18', 'S33'], 'Fixed' : ['DE1234']}
        """
        washable_chars = ',:;./()[]{}<>'
        washed_message = message[:]
        for char in washable_chars:
            washed_message = washed_message.replace(char, ' ')
        washed_message = washed_message.replace('- ', ' ').replace(' -', ' ')
        washed_message = re.sub(r'([A-Z]{1,6}\d+)-(\w+)', r'\1 \2', washed_message)

        msg_stack = washed_message.split()
        msg_stack.reverse()

        actgrp = None
        action_groups = []

        art_ident_pattern = re.compile(r'(?P<art_prefix>%s)(?P<art_num>\d+)' % "|".join(self.valid_artifact_abbrev), re.I)

        while msg_stack:
            word = msg_stack.pop()

            action_word = ACTION_WORD_PATTERN.match(word)
            artifact_word = art_ident_pattern.match(word)

            if not action_word and not artifact_word:
                continue

            if not actgrp:
                actgrp = ActionGrouping()

            if action_word:
                if actgrp.populated(intent='action'):
                    action_groups.append(actgrp)
                    actgrp = ActionGrouping()

                if word not in self.valid_action_words:
                    word = None
                actgrp.setAction(word)

            if artifact_word:
                if actgrp.populated(intent='target'):
                    action_groups.append(actgrp)
                    actgrp = ActionGrouping()
                actgrp.addTarget(word)

        if actgrp:
            action_groups.append(actgrp)

        groups = self.coalesceActionGroups(action_groups)
        # now create our action_groups dictionary to return to caller
        action_groups = {}
        for ag in groups:  # ag --> ActionGroup instance
            if ag.targets:
                #action_groups[ag.action] = ag.targets
                targets = [target for target in ag.targets if ag.action in self.valid_artifact_actions[re.sub(r'\d+$', '', target)] ]
                if targets:
                    action_groups[ag.action] = targets
        return action_groups

    service = extractActionGroups

    def coalesceActionGroups(self, action_groups):
        """
            given a list of ActionGroup instances, produce a list of ActionGroups
            in which each action is unique.
            Multiple instances of an ActionGroup with the same action contribute
            their targets to the single resultant holder of the action.
        """
        min_list = []
        for ag in action_groups:
            uniq = [u for u in min_list if u.action == ag.action]
            if not uniq:
                min_list.append(ag)
            else:
                uniq = uniq.pop()
                uniq.targets.extend(ag.targets)

        return min_list

#####################################################################################################

class ActionGrouping(object):
    def __init__(self):
        self.action = None
        self.targets = []
        self.first_item = None

    def setAction(self, action):
        self.action = action
        if not self.first_item:
            self.first_item = 'action'
            # print "set action to %s , first_item? %s" % (action, True if self.first_item == 'action' else False)

    def addTarget(self, target):
        self.targets.append(target)
        if not self.first_item:
            self.first_item = 'target'

    def populated(self, intent='action'):
        if not self.targets:
            return False
        if not self.action:
            return False

        if intent == 'action':
            return True
        if intent == 'target' and self.first_item == 'target':
            return True

        return False

    def __repr__(self):
        action = self.action
        if self.action:
            action = "'%s'" % self.action
        return "%s: %s" % (action, self.targets)

#####################################################################################################

TEST_ITEMS = \
    [
        ('Completed the work of 12 people in 1 hour',  # non-conforming
         {}),

        ('STP9321 this is not a reference to a Completed DE15 S123 artifact',  # conforming
         {'Completed': ['S123']}),

        ('STP9321 this is not a reference to a Completed DE15 US123 artifact',  # conforming
         {}),

        ('USB-2  not a valid artifact reference',  # non-conforming
         {}),

        ('S123 Completed',  # conforming
         {'Completed': ['S123']}),

        ('DE103 when will this ever be fixed?',  # conforming
         {}),

        ('DE123 Fixed and you can bank on it',  # conforming
         {'Fixed': ['DE123']}),

        ('S123 Completed Fixed TA123 S123',  # conforming (but non-std))
         {'Completed': ['S123']}),

        ('S1234 S3421 a wo0ookies work is never done...',  # conforming
         {}),

        ('TA123 TA345 Fixodent with molars behind',  # conforming
         {}),

        ('Fixed DE1054 4324 lower case chaos KG324',  # conforming
         {'Fixed': ['DE1054']}),

        ('Completed S3245',  # conforming
         {'Completed': ['S3245']}),

        ('Replicated DE123 DE832',  # conforming
         {'Replicated': ['DE123', 'DE832']}),

        ('Fixed DE532 Completed S342 for your peace of mind',  # conforming
         {'Fixed': ['DE532'], 'Completed': ['S342']}),

        ('Replicated DE123 DE832 Defiled S174',  # conforming
         {'Replicated': ['DE123', 'DE832']}),

        ('Replicated DE234 DE329 Barfek S324 S432',  # conforming
         {'Replicated': ['DE234', 'DE329']}),

        ('Completed S432 Fixed DE549 Completed TA617 now go home',  # conforming
         {'Completed': ['S432', 'TA617'], 'Fixed': ['DE549']}),

        ('DE543 Completed DE322 Fixed S235 Completed overnight delivery',  # conforming
         {'Completed': ['S235'], 'Fixed': ['DE322']}),

        ('Fixed - DE322 piece of cake Open: DE333,DE334',  # conforming
        {'Fixed': ['DE322'], 'Open':['DE333', 'DE334']}),

        ('DE322- Fixed piece of cake, Open: DE333,:[DE334]',  # conforming
         {'Fixed': ['DE322'], 'Open': ['DE333', 'DE334']}),
    ]


#####################################################################################################

def test(args):
    artifact_specs = {('S', 'ScheduleState'):['Defined', 'InProgress', 'Implemented', 'Completed'],
                      ('DE', 'State'): ['Open', 'Fixed', 'Replicated', 'Closed'],
                      ('TA', 'State'): ['Completed', 'Vindicated']
                     }

    for zbix, pair in enumerate(TEST_ITEMS):
        ix = zbix + 1
        message, expected = pair
        print("%2d: %s" % (ix, message))
        print("    expected ->   %s" % (repr(expected)))
        extractor = BasicActionsAndArtifactsExtractor(artifact_specs)
        action_targets = extractor.extractActionGroups(message)
        if len(action_targets) == 0 and len(expected) == 0:
            status = 'conforming'
        else:
            if expected and not action_targets:
                status = 'non-conforming'
            elif not expected and action_targets:
                status = 'non-conforming'
            else:
                if expected == action_targets:
                    status = 'conforming'
                else:
                    status = 'non-conforming'

        print("    %s    %s" % (status, repr(action_targets)))
        print("")


#####################################################################################################
#####################################################################################################

if __name__ == '__main__':
    test(sys.argv[1:])