# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import sys,os
import re
import time
import vcsconn.utils.ac_prefixes  as utils
from collections import OrderedDict

from vcsconn.utils.eif_exception import ConfigurationError, OperationalError
from vcsconn.connection import VCSConnection
from vcsconn.agicen_transformer import AgicenTransformer
from vcsconn.utils.auxloader import AuxLoader
#from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT

from pyral import Rally, rallySettings, RallyRESTAPIError


############################################################################################

__version__ = "2.0.5"

VALID_ARTIFACT_PATTERN = None # set after config with artifact prefixes are known

ISO8601_TS_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
VALID_AUTHOR_TRANSFORMS = ['Passthrough', 'FileBasedUserNameLookup', 'UserNameDomainAugmentLookup']
EXTENSION_SPEC_PATTERN  = re.compile(r'^(?P<ext_class>[\w\.]+)\s*\((?P<ext_parm>[^\)]+)\)$')

############################################################################################

class AgileCentralConnection(VCSConnection):

    def __init__(self, config, logger):
        super().__init__(logger)
        self.internalizeConfig(config) 
        self.log.info("Rally WSAPI Version %s" % self.rallyWSAPIVersion())
        self.other_name        = 'Unknown'
        self.other_version     = ""
        self.username_required = True
        self.password_required = True
        self.ac_user           = {}  # keyed by mc.email_address, value pyral User object
        self.gogolofied_users  = {}
        self.prefixes_cache    = []


    def name(self):
        return "AgileCentral"

    def version(self):
        global __version__
        return __version__

    # Rally Web Services version, do not modify unless tested!
    def rallyWSAPIVersion(self):
        return "v2.0"

    def getBackendVersion(self):
        """
            Conform to Connection subclass protocol which requires the version of 
            the system this instance is "connected" to.
        """
        return "Rally WSAPI %s" % self.rallyWSAPIVersion()

    def internalizeConfig(self, config):

        super().internalizeConfig(config)

        server = config.get('Server', 'rally1.rallydev.com')
        if 'http' in server.lower() or '/slm' in server.lower():
            self.log.error(self, "AgileCentral URL should be in the form 'rally1.rallydev.com'")
        self.server = server

        self.url = "https://%s/slm" % server
        self.port = config.get('Port', '443')
        self.apikey          = config.get("APIKey", config.get("API_Key", None))
        self.workspace_name  = config.get("Workspace", None)
        self.project_name    = config.get("Project",   None)  # This gets bled in by the VCSConnector
        self.restapi_debug   = config.get("Debug", False)
        self.restapi_logger  = self.log

        self.proxy = None
        if self.proxy_server:
            self.proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                self.proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
                self.fuzzed_proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, '********', '********', self.proxy_server, self.proxy_port)

        self.transform  = config.get("Transforms", None)
        if not self.transform:
            self.transform = {}
            self.transform['Author'] = 'Passthrough'

        valid_config_items = ['Server', 'Port', 'APIKey', 'API_Key', 'Workspace','Project','Username','Password',
                              'ProxyProtocol', 'ProxyServer','ProxyPort','ProxyUser','ProxyUsername', 'ProxyPassword',
                              'Debug', 'Transforms','UpdateArtifactState','StateExtractorClass']
        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "AgileCentral section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)

        self.update_state    = config.get("UpdateArtifactState", None)
        self.state_extractor = config.get("StateExtractorClass", None)
        self.extension = {}


    def validate(self):
        satisfactory = True

        if self.username_required:
            if not self.username and not self.apikey:
                self.log.error("<Username> is required in the config file")
                satisfactory = False
            #else:
            #    self.log.debug(
            #        '%s - user entry "%s" detected in config file' % (self.__class__.__name__, self.username))

        if self.password_required:
            if not self.password and not self.apikey:
                self.log.error("<Password> is required in the config file")
                satisfactory = False
            #else:
            #    self.log.debug('%s - password entry detected in config file' % self.__class__.__name__)

        if not self.transform['Author'].split('(')[0] in VALID_AUTHOR_TRANSFORMS:
            self.log.info("VCSConnector validation failed: Invalid Author Transform value specified: %s" %
                self.transform['Author'])
            return False
        try:
            self.transformer = AgicenTransformer(self.transform)
            self.transformer.validate()
            self.transformer.loadTransforms()
        except Exception as ex:
            self.log.error(ex)
            return False

        if not self.update_state and self.state_extractor:
            msg = 'StateExtractorClass requires configuration item UpdateArtifactState set to True'
            self.log.error(msg)
            return False
        elif self.update_state and not self.state_extractor:
            msg = 'UpdateArtifactState requires configuration item StateExtractorClass set to a valid class name'
            self.log.error(msg)
            return False

        if self.update_state and self.state_extractor:
            self.artifacts_metadata = self.getMetadata()
            self.aux = AuxLoader()
            self.aux.loadFacility('Extension', self.state_extractor)
            self.extension['StateExtractor'] = self.aux.getFacility('Extension', self.state_extractor)(self.artifacts_metadata)

            if not self.extension['StateExtractor']:
                msg = 'cannot load StateExtractorClass specified in the config file'
                self.log.error(msg)
                return False
        return satisfactory

    def setSourceIdentification(self, other_name, other_version):
        self.other_name = other_name
        self.other_version  = other_version

    def get_custom_headers(self):
        custom_headers =  {}
        #custom_headers['name']    = "AgileCentral VCS Connector for GitHub"
        custom_headers['name'] = "AgileCentral VCS Connector for %s" % self.other_name
        custom_headers['vendor']  = "CA Technologies"
        custom_headers['version'] = self.version()
        if self.other_version:
            spoke_versions = "%s - %s " % (self.version(), self.other_version)
            custom_headers['version'] = spoke_versions
        return custom_headers


    def connect(self):
        if self.proxy:
            self.log.info("Proxy for AgileCentral connection:  %s" % self.fuzzed_proxy)
            os.environ['HTTPS_PROXY'] = self.proxy

        self.log.info("Connecting to AgileCentral")
        custom_headers = self.get_custom_headers()

        try:
            self.agicen = Rally(self.server, user=self.username, password=self.password, apikey=self.apikey,
                                workspace=self.workspace_name, project=self.project_name,
                                version=self.rallyWSAPIVersion(), headers=custom_headers,
                                server_ping=False, isolated_workspace=True,
                                logger=self.restapi_logger, warn=False, debug=True)

            if self.restapi_debug:
                self.agicen.enableLogging('agicen_builds.log')
        except Exception as msg:
            raise ConfigurationError("Unable to connect to Agile Central at %s: %s" % \
                                         (self.server, msg))
        self.log.info("Connected to Agile Central server: %s" % self.server)    

        # verify the given workspace name exists
        all_workspaces = self.agicen.getWorkspaces()

        valid = [wksp for wksp in all_workspaces if wksp.Name == self.workspace_name]
        if not valid:
            problem = "Specified Workspace: '%s' not in list of workspaces " + \
                      "available for your credentials as user: %s" 
            raise ConfigurationError(problem % (self.workspace_name, self.username))
        self.log.info("    Workspace: %s" % self.workspace_name)
        self.log.info("    Project  : %s" % self.project_name)
        wksp = self.agicen.getWorkspace()
        prjt  = self.agicen.getProject()
        self.workspace_ref = wksp.ref
        self.project_ref   = prjt.ref

        # find all of the Projects under the AgileCentral_Project
        response = self.agicen.get('Project', fetch='Name', workspace=self.workspace_name,
                                   project=self.project_name,
                                   projectScopeDown=True,
                                   pagesize=200)
        if response.errors or response.resultCount == 0:
            raise ConfigurationError('Unable to locate a Project with the name: %s in the target Workspace' % self.project_name)

        # detect any duplicate project names
        self.project_bucket = {}
        project_names = [proj.Name for proj in response]
        for project_name in project_names:
            if project_name not in self.project_bucket:
                self.project_bucket[project_name] = 0
            self.project_bucket[project_name] += 1
        self.duplicated_project_names = [p for p,c in self.project_bucket.items() if c != 1]

        self.log.info("    %d sub-projects" % len(project_names))
        self.prefixes_cache = utils.get_all_prefixes(self.agicen)

        return True


    def disconnect(self):
        """
            Just reset our agicen instance variable to None
        """
        self.agicen = None


    def set_integration_header(self, header_info):
        self.integration_name    = header_info['name']
        self.integration_vendor  = header_info['vendor']
        self.integration_version = header_info['version']
        if 'other_version' in header_info:
            self.other_version = header_info['other_version']


    def getRecentCommits(self, ref_time):
        """
            Obtain all Changesets created in Agile Central at or after the ref_time parameter which is an instance of TimeStamp
        """
        self.log.info("Detecting recently added Agile Central Changesets")
        selectors = ['CommitTimestamp >= %s' % ref_time.asISOString()]
        log_msg = '   recent Changesets query: %s' %  ' and '.join(selectors)
        self.log.info(log_msg)
        changesets = []

        response = self._retrieveChangesets(selectors)
        log_msg = "  %d recently added Agile Central Changesets"
        self.log.info(log_msg % response.resultCount)

        for changeset in response:
            changesets.append(changeset)
        return changesets


    def _retrieveChangesets(self, selectors):
        chgset_fields = "ObjectID,SCMRepository,Name,Revision,CommitTimestamp,Author,Changes,CreationDate"
        try:
            response = self.agicen.get('Changeset',
                                       fetch=chgset_fields,
                                       query=selectors,
                                       workspace=self.workspace_name,
                                       project=None,
                                       projectScopeDown=False,
                                       projectScopeUp=False,
                                       order="CommitTimestamp",
                                       pagesize=1000
                                      )
        except Exception as msg:
            problem = ''
            excp_type, excp_value, tb = sys.exc_info()
            mo = re.search(r"'(?P<ex_name>.+)'", str(excp_type))
            if mo:
                excp_type = mo.group('ex_name').replace('exceptions.', '')
                problem = '%s: %s\n' % (excp_type, str(excp_value))
            raise OperationalError(problem)

        # Examine the response to see if there is any content for the 'Errors' or 'Warnings' keys.
        # and raise an Exception in that case.
        if response.errors:
            raise Exception(response.errors[0][:80])
        if response.warnings:
            raise Exception(response.warnings[0][:80])

        return response

    def retrieveChangeset(self, sha):
        query = 'Revision = %s' % sha
        response = self.agicen.get('Changeset', fetch='ObjectID', query=query,
                                    workspace=self.workspace_name,project=None)
        if response.resultCount > 0:
            return response.next()
        return None

    def postCommits(self, unrecorded_commits, append_vcs_committer):
        # append_vcs_committer is a boolean parm to be used in the call to createChangesets
        changesets = OrderedDict()
        repo_names = list(unrecorded_commits.keys())
        for repo_name in repo_names:
            scm_repo = self.ensureSCMRepositoryExists(repo_name, vcs_type='git')
            changesets[repo_name] = self.createChangesets(scm_repo, repo_name, unrecorded_commits, append_vcs_committer)
        for email_address in self.gogolofied_users:
            self.log.info('Did not find an AgileCentral user with  username: %s' % email_address)

        return changesets

    def postPullRequests(self, unrecorded_pull_requests):
        # order the unrecorded_pull_requests by oldest to youngest
        unrecorded_pull_requests = reversed(unrecorded_pull_requests)
        posted_pull_requests, skipped_pull_requests = self.createPullRequests(unrecorded_pull_requests)
        self.log.info("PullRequests posted : %d" % len(posted_pull_requests))
        self.log.info("PullRequests skipped: %d  (no AC Artifacts was mentioned in commits)" % len(skipped_pull_requests))
        return posted_pull_requests


    def ensureSCMRepositoryExists(self, repo_name, vcs_type):
        """
            See if the repo_name exists in AC, and if not create it.
        """
        criteria = '(Name = "{0}")'.format(repo_name)
        response = self.agicen.get('SCMRepository', fetch="Name,ObjectID", query = criteria, project=None)
        if response.resultCount:
            scm_repos = [item for item in response]
            exact_matches = [scm_repo for scm_repo in scm_repos if scm_repo.Name.lower() == repo_name.lower()]
            if exact_matches:
                return exact_matches[0]

        scm_repo_payload = {'Name': repo_name, 'SCMType': vcs_type}
        try:
            scm_repo = self.agicen.create('SCMRepository', scm_repo_payload)
            self.log.info("Created SCMRepository %s (%s)" % (scm_repo.Name, scm_repo.ObjectID))
        except RallyRESTAPIError as msg:
            raise OperationalError("Could not create SCMRepository %s" % repo_name)

        return scm_repo

    def getAgileCentralUser(self, email):
        if email in self.ac_user:
            return self.ac_user[email]
        query = '(UserName = %s)' %email
        response = self.agicen.get('User', fetch='UserName', query=query, pagesize=1, limit=1)
        user = None
        if response.resultCount == 0:
            if email not in self.gogolofied_users:
                self.gogolofied_users[email] = None
        else:
            user = response.next()
            self.ac_user[email] = user
        return user

    def changesetExists(self, scmrepo_ref, sha, timestamp):

        query = ['SCMRepository = %s' % scmrepo_ref, 'Revision = %s' % sha, 'CommitTimestamp = %s' % timestamp]

        try:
            response = self.agicen.get('Changeset', query=query, project=None)
            return True if response.resultCount else False
        except:
            return False

    def createChangesets(self, scm_repo, repo_name, missing_changesets, append_vcs_committer):
        self.log.debug("Creating Changesets for SCMRepository: %s" % scm_repo.Name)
        ac_changesets = []

        all_mentioned_artifact_fids = {mc.rev_num: self.parseForArtifacts(mc.message) for mc in missing_changesets[repo_name]}
        valid_artifact_fids         = self.validatedArtifacts(all_mentioned_artifact_fids)
        for mc in missing_changesets[repo_name]:
            commit_message = mc.message[:]
            if append_vcs_committer:
                commit_message += ("\n - committed by: |%s|" % mc.committer)
            ac_user = self.getAgileCentralUser(self.transformer.transform('Author', mc) or mc.committer_email)
            valid_artifacts = valid_artifact_fids[mc.rev_num]
            changeset_payload = {
                'SCMRepository'   : scm_repo.ref,
                'Revision'        : mc.rev_num,
                'CommitTimestamp' : mc.commit_timestamp.asISOString(),
                'Message'         : commit_message[:3999],  # AC WSAPI only allows 4000 chars for Message
                'Uri'             : mc.path,
                'Artifacts'       : valid_artifacts,
                'Author'          : ac_user.ref if ac_user else None
            }

            # consider the need for detecting a duplicate Changeset by sha here before actually creating...
            # create a self.agicen.get query condition with SCMRepository, Revision, CommitTimestamp
            # if there already is a Changeset with those values, bypass posting one now
            # otherwise go ahead and create the Changeset
            if self.changesetExists(scm_repo.ref, mc.rev_num, mc.commit_timestamp):
                continue
            changeset = self.agicen.create('Changeset', changeset_payload)
            cs_author = changeset.Author.UserName if changeset.Author else ''
            self.log.debug("    Created Changeset for SHA: %s   CommitTimestamp: %s   Author: %s" % \
                               (changeset.Revision, changeset.CommitTimestamp.split('.000')[0] + 'Z', cs_author))
            ac_changesets.append(changeset)

            # should we be executing our UpdateArtifactState extension here?
            if self.update_state and self.extension['StateExtractor']:
                action_targets = self.extension['StateExtractor'].service(mc.message)
                if action_targets:
                    self.updateState(action_targets, valid_artifacts)
        #most_recent_per_repo = ac_changesets[-1]
        return ac_changesets


    def parseForArtifacts(self, commit_message):
        prefixes = [prefix for item in self.prefixes_cache for prefix in item.values()]
        fid_pattern = r'\b((%s)\d+)\b' % '|'.join(prefixes)
        result = re.findall(fid_pattern, commit_message, re.IGNORECASE)
        return [item[0].upper() for item in result]


    def validatedArtifacts(self, commit_fid):
        # commit_fid is a dict
        # commit_fid = {commitID: [S123,DE123], ...}
        try:
            fids = list(set([fid for fids in commit_fid.values() for fid in fids]))
        except Exception as msg:
            problem = "Cannot get a list of formatted ids of artifacts in the commit messages, in validatedArtfacts"
            raise OperationalError(problem)
        query = self.makeOrQuery("FormattedID", fids)
        response = self.agicen.get('Artifact', fetch="FormattedID,ObjectID", query=query, project=None, pagesize=200, start=1)
        found_arts = [art for art in response]
        va = {}
        for ident in commit_fid.keys():
            va[ident] = []
            matches = [found_art for found_art in found_arts if found_art.FormattedID in commit_fid[ident]]
            if matches:
                va[ident] = matches
        return va

    def makeOrQuery(self,field, values):
        if not values:
            return None

        values = list(values)
        query = '(%s = "%s")' %(field, values[0])
        if len(values) == 1:
            return query
        else:
            values.pop(0)
            for v in values:
                query = '(%s OR (%s = "%s"))' % (query, field, v)
            return query


    def matchTypeToField(self, artifact_type):
        return{
            'Defect'  : 'State',
            'Task'    : 'State',
            'HierarchicalRequirement':'ScheduleState'
        }[artifact_type]

    def getArtifactTypesInfo(self, element_name):
        fields = "ElementName,IDPrefix"
        query = 'ElementName = "%s"' % element_name
        art_type = self.agicen.get('TypeDefinition', fetch=fields, query=query, order="Ordinal", instance=True)
        field = self.matchTypeToField(art_type.ElementName)
        av = self.agicen.getAllowedValues(art_type.ElementName, field)
        return (art_type.IDPrefix, field), av

    def getMetadata(self):
        art_types = ['HierarchicalRequirement', 'Defect', 'Task']
        metadata = {}
        for a_type in art_types:
            key, value = self.getArtifactTypesInfo(a_type)
            metadata[key] = value
        return metadata

    def updateState(self, action_targets, artifacts):
        """
        """
        # action_targets looks like this {'In-Progress': ['US210']}
        actions = [key for key in action_targets if key != None]
        self.log.debug("_updateArtifacts has %d actions" % len(actions))
        for action in actions:
            targets = action_targets[action]
            self.log.debug("_updateArtifacts has %d artifacts to %s" % (len(targets), action))
            for target in targets:
                self.log.debug("_updateArtifacts target: %s" % target)
                target_artifacts = [art for art in artifacts if art.FormattedID == target]
                if not target_artifacts:
                    self.log.debug("   there are no target artifacts")
                    continue
                for artifact in target_artifacts:
                    #print (artifact.FormattedID)
                    _type = artifact._type
                    fid   = artifact.FormattedID
                    field_to_update = self.matchTypeToField(_type)
                    artifact_data = { "FormattedID"   : fid,
                                      field_to_update : [item for item in action_targets.keys()][0]
                                    }
                    try:
                        self.agicen.update(_type, artifact_data)
                        # "Errors": ["Could not set value for Schedule State directly
                    except Exception as ex:
                        self.log.error(ex)

    def _retrievePullRequests(self, selectors):
        pr_fields = "ObjectID,ExternalID,ExternalFormattedID,Name,CreationDate,Artifact,FormattedID"
        try:
            response = self.agicen.get('PullRequest',
                                       fetch=pr_fields,
                                       query=selectors,
                                       workspace=self.workspace_name,
                                       project=None,
                                       projectScopeDown=False,
                                       projectScopeUp=False,
                                       #order="CreationDate", # Could not find attribute "CreationDate" on type PullRequest
                                       pagesize=1000
                                      )
        except Exception as msg:
            problem = ''
            excp_type, excp_value, tb = sys.exc_info()
            mo = re.search(r"'(?P<ex_name>.+)'", str(excp_type))
            if mo:
                excp_type = mo.group('ex_name').replace('exceptions.', '')
                problem = '%s: %s\n' % (excp_type, str(excp_value))
            raise OperationalError(problem)

        # Examine the response to see if there is any content for the 'Errors' or 'Warnings' keys.
        if response.errors:
            raise RallyRESTAPIError(response.errors[0][:80])
        if response.warnings:
            raise RallyRESTAPIError(response.warnings[0][:80])

        return response

    def getRecentPullRequests(self, ref_time):
        """
           use our self.agicen instance to issue a get request for PullRequest items whose CreationDate is >= ref_time
        """
        self.log.info("Detecting recently added Agile Central PullRequests")
        selectors = ['CreationDate >= %s' % ref_time.asISOString()]
        log_msg = '   recent PullRequests query: %s' % ' and '.join(selectors)
        self.log.info(log_msg)
        pull_requests = []

        response = self._retrievePullRequests(selectors)
        log_msg = "  %d recently added Agile Central PullRequests"
        self.log.info(log_msg % response.resultCount)

        for pr in response:
            pull_requests.append(pr)
        return pull_requests

    def pullRequestExists(self, externalid, artifact):
        query = ['ExternalID = %s' % externalid, 'Artifact = %s' % artifact.ref]
        try:
            response = self.agicen.get('PullRequest', query=query, project=None)
            return True if response.resultCount else False
        except:
            return False


    def constructPullRequestPayload(self, pull_request, artifact):
        pr = pull_request
        payload = {
                   'ExternalID'          : pr.ident,
                   'ExternalFormattedId' : pr.number,
                   'Name'                : pr.title,
                   'Description'         : pr.description,
                   'Url'                 : pr.url,
                   'Artifact'            : artifact.ref
                  }

        return payload

    def identifyValidArtifacts(self, pull_request):
        """
            Look for valid artifact FormattedIDs in the commit messages AND in the
            title for the PullRequest.  Shove the PullRequest title value in to the
            list of commit messages so that is also considered.
        """
        commit_messages = pull_request.extractCommitMessages()
        commit_messages.insert(0, pull_request.title)
        related_fids = [{pull_request.ident: self.parseForArtifacts(commit_message)} for commit_message in
                        commit_messages]
        # wash out any items in related_fids that have an empty list for fids
        related_fids = [item for item in related_fids for value in item.values() if value]
        fid_list = []
        [fid_list.extend(element.values()) for element in related_fids]
        fids = [item for sublist in fid_list for item in sublist]
        if not fids:
            return []
        artifacts = self.validatedArtifacts({pull_request.ident: fids})[pull_request.ident]
        return artifacts


    def createPullRequests(self, unrecorded_pull_requests):
        ac_pull_requests      = []
        skipped_pull_requests = []
        for pull_request in unrecorded_pull_requests:
            artifacts = self.identifyValidArtifacts(pull_request)
            if not artifacts:
                skipped_pull_requests.append(pull_request.ident)
            for artifact in artifacts:
                if self.pullRequestExists(pull_request.ident, artifact):
                    continue
                pr_payload = self.constructPullRequestPayload(pull_request, artifact)

                try:
                    ac_pull_request = self.agicen.create('PullRequest', pr_payload)
                    self.log.debug("    Created PullRequest in AgileCentral for  %s (%s)" % (ac_pull_request.Url, artifact.FormattedID))
                    ac_pull_requests.append(ac_pull_request)
                except RallyRESTAPIError as ex:
                    raise OperationalError("Could not create PullRequest  %s" % ex.args[0])

        return ac_pull_requests, skipped_pull_requests
