# Copyright 2017, 2018 CA Technologies. All Rights Reserved.


import os
import re
import copy

import requests
from vcsconn.connection import VCSConnection
from vcsconn.utils.eif_exception import ConfigurationError, OperationalError

from vcsconn.bitbucket_helper import BitbucketHelper
from vcsconn.bitbucket_pullrequest import  BitbucketPullRequest
from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT, READABLE_FORMAT

############################################################################################

class BitbucketConnection(VCSConnection):

    BITBUCKET_CONN_VERSION = '2.0.5'

    def __init__(self, config, logger):
        super().__init__(logger)
        self.server         = None
        self.projectkey     = None
        self.internalizeConfig(config)
        self.backend_version     = None
        self.processable_repos   = []
        self.bb_helper           = None
        return

    def name(self):
        return "Bitbucket"

    def version(self):
        """
            Return the version of BitbucketConnection class
        """
        return self.BITBUCKET_CONN_VERSION


    def connect(self):
        self.log.info("Connecting to Bitbucket...")
        self.backend_version = None
        problem = ""
        if self.cloud:
            if self.config.get('UserSlug', False) and self.config.get('TeamSlug', False):
                problem = 'UserSlug and TeamSlug are both present in the configuration file. Only one of them can be present.'
                raise ConfigurationError(problem)
            if not self.config.get('UserSlug', False) and not self.config.get('TeamSlug', False):
                problem = 'UserSlug and TeamSlug are missing from the configuration file. At least one of them must be present.'
                raise ConfigurationError(problem)
            self.base_url = "%s://%s:%s/2.0" % (self.protocol, self.server, self.port)
            if self.config.get('UserSlug', False):
                uri = "%s/users/%s" % (self.base_url, self.userslug)
                try:
                    headers = {'Content-Type': 'application/json'}
                    response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
                    result = response.json()
                    if result['username'].lower() != self.userslug.lower():
                        problem = "UserSlug provided in the configuration file is invalid."
                        raise ConfigurationError(problem)
                    else:
                        self.backend_version = '2.0'
                        self.log.info("Connected to Bitbucket Cloud at %s" % self.server)
                except Exception as e:
                    problem = "Unable to connect to Bitbucket Cloud at %s for UserSlug. %s" % (uri, e)
                    raise ConfigurationError(problem)
            if self.config.get('TeamSlug', False):
                uri = "%s/teams/%s" % (self.base_url, self.teamslug)
                try:
                    headers = {'Content-Type': 'application/json'}
                    response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
                    result = response.json()
                    if result['username'].lower() != self.teamslug.lower():
                        problem = "TeamSlug provided in the configuration file is invalid."
                        raise ConfigurationError(problem)
                    else:
                        self.backend_version = '2.0'
                        self.log.info("Connected to Bitbucket Cloud at %s" % self.server)
                except Exception as e:
                    problem = "Unable to connect to Bitbucket Cloud at %s for TeamSlug. %s" % (uri, e)
                    raise ConfigurationError(problem)

        else:
            if not self.config.get('ProjectKey'):
                problem = "ProjectKey is missing from the configuration file."
            else:
                if not self.custom_prefix:
                    self.base_url = "%s://%s:%s/rest/api/1.0" % (self.protocol, self.server, self.port)
                else:
                    self.base_url = "%s://%s:%s/%s/rest/api/1.0" % (self.protocol, self.server, self.port, self.custom_prefix.strip('/'))
                uri = "%s/projects/%s" % (self.base_url, self.projectkey)
                try:
                    headers = {'Content-Type': 'application/json'}
                    response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
                    result = response.json()
                    if result['key'] != self.projectkey:
                        problem = "ProjectKey in the configuration file is invalid."
                        raise ConfigurationError(problem)
                    else:
                        self.backend_version = '1.0'
                        self.log.info("Connected to Bitbucket Server at %s" % self.server)
                except Exception as e:
                    problem = "Unable to connect to Bitbucket Server at %s. %s" % (uri, e)
                    raise ConfigurationError(problem)

        if not self.backend_version:
            self.log.error(problem)
            return False
        return self.backend_version


    def internalizeConfig(self, config):
        super().internalizeConfig(config)
        self._setConnectionDefaults(config)

        self.protocol         = config.get('Protocol', 'http')
        self.server           = config.get('Server',    None)
        self.port             = config.get('Port',      7990)
        self.custom_prefix    = config.get('Prefix',    None)
        #self.username         = config.get('Username',  None) # taken care of in Connection superclass method
        #self.password         = config.get('Password',  None) # ditto
        self.cert_file        = config.get('CertFile',  None)
        if self.cloud:
            self.userslug     = config.get('UserSlug', "")
            self.teamslug     = config.get('TeamSlug', "")
        else:
            self.userslug     = ""
            self.teamslug     = ""
        self.projectkey       = config.get('ProjectKey', None)
        self.backend_version  = None
        self.debug            = config.get('Debug'     , False)
        self.repo_spec = config.get('Repository', None)

        self.http_proxy = {}
        if self.proxy_server:
            proxy = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                proxy = "%s://%s:%s@%s:%s" % (
                    self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
            self.http_proxy = {self.protocol: proxy}
            self.log.info("Proxy for Bitbucket connection:  %s" % proxy)

        # Fixup things we need
        if self.server.__class__.__name__ == 'str':
            self.server = re.sub(r'^https?:\/', '', self.server)

        self.port   = str(self.port)
        if self.debug and self.debug.lower() in ['false', 'no', 'off', '0']:
            self.debug = False
        if self.debug and self.debug.lower() in ['true', 'yes', 'on', '1']:
            self.debug = True

        # Not all fields are required for cloud
        self.port_required        = True
        self.projectkey_required  = True
        if self.cloud:
            self.port_required        = False
            self.projectkey_required  = False

        if self.username and self.password:
            self.auth = (self.username, self.password)
        else:
            return False

        if self.cert_file:
            os.environ['REQUESTS_CA_BUNDLE'] = self.cert_file  # '/etc/ssl/certs/cacert.pem' or other location...

    def _setConnectionDefaults(self, config):
        DEFAULT_BITBUCKET_SERVER_PORT  = 7990
        DEFAULT_BITBUCKET_PORT_FOR_SSL =  443
        BITBUCKET_CLOUD_SITE           = 'api.bitbucket.org'
        if not config.get('Server', None) or config.get('Server', None) == BITBUCKET_CLOUD_SITE:
            self.cloud = True
            config['Protocol'] = 'https'
            config['Server'  ] = BITBUCKET_CLOUD_SITE
            config['Port'    ] = DEFAULT_BITBUCKET_PORT_FOR_SSL
        elif config.get('Server', None): # config.get('Server', None) != BITBUCKET_CLOUD_SITE
            self.cloud = False
            if not config.get('Protocol', None):
                config['Protocol'] = 'http'
            if not config.get('Port', None):
                config['Port'] = DEFAULT_BITBUCKET_SERVER_PORT


    def disconnect(self):
        pass

    def validate(self):
        satisfactory = True
        self.bb_helper = BitbucketHelper(self.username, self.password, self.log, self.cloud, self.http_proxy)
        accessible_repos = []
        if self.cloud:
            if self.userslug and self.teamslug:
                self.log.error("'UserSlug:' and 'TeamSlug:' are both present in the Bitbucket section of the config file.")
                return False
            if not self.userslug and not self.teamslug:
                self.log.error("'UserSlug:' and/or 'TeamSlug:' is required in the Bitbucket section of the config file")
                return False
            if self.userslug:
                self.resource_slug = self.userslug
                self.log.debug("Bitbucket - 'UserSlug:' entry '%s' detected in the Bitbucket section of the config file" % (self.userslug))
            if self.teamslug:
                self.resource_slug = self.teamslug
                self.log.debug("Bitbucket - 'TeamSlug:' entry '%s' detected in the Bitbucket section of the config file" % (self.teamslug))

            repos_url = "%s/repositories/%s" % (self.base_url, self.resource_slug)
            accessible_repos = self.obtainBitbucketAccessibleRepositories(repos_url)
        else:
            if not self.projectkey:
                self.log.error("'ProjectKey:' is required in the Bitbucket section of the config file")
                return False
            self.log.debug("Bitbucket - 'ProjectKey:' entry '%s' detected in the Bitbucket section of the config file" % (self.projectkey))
            project_repos_url = "%s/projects/%s/repos" % (self.base_url, self.projectkey)
            accessible_repos = self.obtainBitbucketAccessibleRepositories(project_repos_url)
            self.log.info("  Bitbucket connect attempt using uri=%s" % project_repos_url)

        if not self.repo_spec or len(self.repo_spec) == 0:
            processable_repos = accessible_repos
        else:
            processable_repos = accessible_repos
            if 'Include' in self.repo_spec:
                processable_repos = self.qualifyRepos(accessible_repos, self.repo_spec['Include'], True)
                if not processable_repos:
                    self.log.error("All repos specified in your config are not found among the accessible repos")
                    return False

            if 'Exclude' in self.repo_spec:
                prior_repos = processable_repos[:]
                processable_repos = self.qualifyRepos(prior_repos, self.repo_spec['Exclude'], False)
                excluded_repos = [repo for repo in prior_repos if repo not in processable_repos]
                self.log.debug("These repos were excluded based on your Repositories -> Exclude criteria: %s" % ", ".join(excluded_repos))

        self.processable_repos = processable_repos
        if not processable_repos:
            self.log.error("There are no repositories that meet your criteria")
            satisfactory = False
        else:
            self.log.debug("These repositories will be processed: %s" % ", ".join(processable_repos))
        return satisfactory

    def obtainBitbucketAccessibleRepositories(self, target_url):
        self.log.info("  Bitbucket connect attempt using uri=%s" % target_url)
        repos = self.bb_helper.getAllRepos(target_url)
        accessible_repos = [repo['name'] for repo in repos for key in repo.keys() if key == 'name']

        return accessible_repos


    def qualifyRepos(self, repo_names, qualifications, include):

        qualification_items = re.split(r',\s*', qualifications)
        amended_repo_names = []
        if not include:
            amended_repo_names = repo_names[:]
        for repo_name in repo_names:

            match = False
            for qual in qualification_items:
                # if qual has no '*' char do a simple string compare and use repo that logic result
                # otherwise change the shell wildcard char '*' into a regex wild card '.*' and use the regex result
                if '*' not in qual:
                    if repo_name.upper() == qual.upper():
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
                else:
                    qual_regex = '^%s' % qual.replace('*', '.*')
                    if re.search(qual_regex, repo_name, re.I): # re.I makes repo name case insensitive
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
        return amended_repo_names


    def getRecentCommits(self, ref_time, last_commit_sha=None):
        # https://developer.atlassian.com/static/rest/bitbucket-server/4.14.4/bitbucket-rest.html#idm46478324755072
        """
            Obtain all Changesets created in Bitbucket at or after the ref_time parameter which is a Time instance
        """
        all_commits = {}  # will be keyed by repo name

        pending_operation = "Detecting recently added Bitbucket Changesets (added on or after %s)"
        self.log.info(pending_operation % ref_time.asLogStamp())
        for repo_slug in self.processable_repos:
            all_commits[repo_slug] = self.getRecentRepositoryCommits(repo_slug, ref_time)
        return all_commits


    def getRecentRepositoryCommits(self, repo_slug, ref_time):
        chrono_ordered_changesets = []
        bb_commit_url_prefix = ''
        bburl       = ''
        # self.base_url in cloud: 'https://api.bitbucket.org:443/2.0'
        # bburl example: 'https://api.bitbucket.org:443/2.0/repositories/TechServicesBitbucket/evinrude'
        # a good URI:    https://bitbucket.org/loriteam1/LoriGitRepo1/commits/065f8e82c6962a0efab08cb63af6eef2046f6eb7
        if self.cloud:
            bb_commit_url_prefix = "%s/%s/%s" % ('https://bitbucket.org', self.resource_slug, repo_slug)
            bburl = "%s/repositories/%s/%s" % (self.base_url, self.resource_slug, repo_slug)
        else:
            #bb_commit_url_prefix = "%s/%s/%s" % (self.base_url, self.projectkey, repo_slug)
            bbs_instance = "%s://%s:%s" % (self.protocol, self.server, self.port)
            bb_commit_url_prefix = "%s/projects/%s/repos/%s" % (bbs_instance, self.projectkey, repo_slug)
            bburl = "%s/projects/%s/repos/%s" % (self.base_url, self.projectkey, repo_slug)
        branches = self.bb_helper.getAllBranches(bburl)
        if not branches:
            return chrono_ordered_changesets
        commits = {}
        #  Get all commits on master branch & and build a list of sha's
        sham = {}  # Sha Masters
        shab = {}  # For branches, all Changesets we care about

        commits['master'], master_sha = self.bb_helper.getBranchCommitQualifiers(bburl, 'master', ref_time, None)
        if 'master' not in commits and 'does not exist' in master_sha:
            return []

        for commit in commits['master']:
            commit.branch = 'master'
            sham[commit.ident] = commit

        stop_sha = master_sha
        other_branches = copy.copy(branches)
        if 'master' in other_branches:
            del other_branches['master']

        #other_branches = [branch for branch in branches if list(branch.keys())[0] != 'master']
        for branch_name, head_sha in other_branches.items():
            #branch_name = list(branch.keys())[0]
            commits[branch_name], junk_sha = self.bb_helper.getBranchCommitQualifiers(bburl, branch_name, ref_time, stop_sha)
            for commit in commits[branch_name]:
                if commit.ident not in sham:
                    commit.branch = branch_name
                    shab[commit.ident] = commit

        # coalesce sham and shab dictionary values into a single array ordered by commit timestamp values
        changesets = list(sham.values()) + list(shab.values())
        chrono_ordered_changesets = sorted(changesets, key=lambda x: x.commit_timestamp)
        for changeset in chrono_ordered_changesets:
            changeset.setPath(bb_commit_url_prefix)

        return chrono_ordered_changesets


    def getRecentPullRequests(self, ref_time):
        """
            Select only those PullRequests whose CreatedAt date is >= ref_time_iso.
            iterate over the self.processable_repos to arrive at a single list of Pull Requests.
            We don't care/mind that the PullRequests are not organized into a struct keyed by repo_slug.
        """
        processable_prs = []
        for repo_slug in self.processable_repos:
            if self.cloud:
                repo_url = "%s/repositories/%s/%s"   % (self.base_url, self.resource_slug, repo_slug)
            else:
                repo_url = "%s/projects/%s/repos/%s" % (self.base_url, self.projectkey, repo_slug)
            # http://koljo03-s4576.ca.com:7990/rest/api/1.0/projects/GUN1/repos/rocket-man
            pull_requests = self.bb_helper.getRecentPullRequests(repo_url, ref_time)
            chrono_ordered_prs = sorted(pull_requests, key=lambda x: x.creation_timestamp)
            processable_prs.extend(chrono_ordered_prs)

        return processable_prs

##############################################################################################
