# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import inspect
import requests
from vcsconn.utils.eif_exception import ConfigurationError, OperationalError
from vcsconn.bitbucket_changeset import BitbucketChangeset
from vcsconn.bitbucket_pullrequest import BitbucketPullRequest
from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT
from vcsconn.utils.klog     import getCaller

class BitbucketHelper:

    STD_PAGESIZE = 100
    REPOS_PAGESIZE = 1000

    def __init__(self, username, password, logger, cloud, proxy):
        self.username = username
        self.password = password
        self.log      = logger
        self.cloud    = cloud
        self.http_proxy = proxy

        if self.username and self.password:
            self.auth = (self.username, self.password)

    def getFromBitbucket(self, uri):  # internal helper method that makes requests to Bitbucket Server REST API
        result = None
        stack_item = inspect.stack()[1]  # assume caller of interest is stack_levels down in the call stack
        frame, filename, line_no, code_unit = stack_item[:4]
        caller = getCaller(frame)
        self.log.debug("Bitbucket request URL on behalf of %s: %s" % (caller, uri))
        try:
            headers = {'Content-Type': 'application/json'}
            response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
            result = response.json()
        except Exception as exc:
            actual_error = exc.args[0]
            problem = "Unable to retrieve information from Bitbucket. %s" % actual_error
            raise OperationalError(problem)

        if self.cloud and 'error' in result:
            msg = result['error']['message']
            raise OperationalError(msg)
        if not self.cloud and 'errors' in result:
            msg = result['errors'][0]['message']
            raise OperationalError(msg)

        return result

    def getAllRepos(self, repos_url):
        # e.g http://koljo03-s4576.ca.com:7990/rest/api/1.0/projects/GUN1/repos or
        #     https://api.bitbucket.org/2.0/repositories/TechServicesBitbucket

        repos_url = repos_url + '?limit=%s' % BitbucketHelper.REPOS_PAGESIZE
        # below is an intentionally invalid url:  (for dev/test purposes only)
        #repos_url =  repos_url +'/notok' + '?limit=%s' % BitbucketHelper.REPOS_PAGESIZE
        try:
            result = self.getFromBitbucket(repos_url)
        except Exception as ex:
            ex.report()
            return []
        return result['values']


    # def getProject(self, project_url):
    #     result = self.getFromBitbucket(project_url)
    #     self.checkForBitbucketErrors(project_url, result)
    #     #return result['values'] ??
    #     return result

    def getAllBranches(self, baseurl):
        # Get info about all the branches in the repo
        refs_fmt = '%s/refs/branches' if self.cloud else '%s/branches'
        branches_url = refs_fmt % baseurl
        try:
            result = self.getFromBitbucket(branches_url)
        except Exception as ex:
            ex.report()
            return {}

        if self.cloud:
            branch_data = [{item['name']     : item['target']['hash']}  for item in result['values']]
        else:
            branch_data = [{item['displayId']: item['latestChangeset']} for item in result['values']]

        branch_data = {branch_name: commit_hash for bd in branch_data for branch_name, commit_hash in bd.items()}
        return branch_data

    def getBranchCommitQualifiers(self, bburl, branch_name, ref_time, start_sha):
        commits_url = "%s/commits?until=refs/heads/%s" % (bburl, branch_name)
        commits = []
        bb_commits = None
        start = 0
        page_num = 0
        stop_sha = ''
        ref_time_met = False

        if branch_name != 'master':
           commits_url += "&since=%s" % start_sha

        errors = []
        done = False
        while not done:
            page_num += 1
            paged_url = "%s&start=%s&limit=%s" % (commits_url, start, self.STD_PAGESIZE)
            try:
                result = self.getFromBitbucket(paged_url)
            except Exception as ex:
                ex.report()
                return ([],'')
            bb_commits = result
            if 'errors' in bb_commits or 'values' not in bb_commits:
                if 'errors' in bb_commits and 'message' in bb_commits['errors'][0]:
                    errors.append(bb_commits['errors'][0]['message'])
                ref_time_met = True
                done = True
                break

            if self.cloud:
                if 'next' not in bb_commits:
                    ref_time_met = True
                else:
                    commits_url = bb_commits['next']
            else:
                start = bb_commits['nextPageStart']
            for commit in bb_commits['values']:
                if self.cloud:
                    commit_ts = TimeStamp.fromFormattedString(commit['date'][:-6] + 'Z', ISO_FORMAT)
                else:
                    commit_ts = TimeStamp.fromEpochSeconds(commit['authorTimestamp'] / 1000)
                if commit_ts.epoch_seconds >= ref_time.epoch_seconds:
                    commits.append(BitbucketChangeset(commit))
                else:
                    if branch_name == 'master':
                        stop_sha = commit.get('id','hash')
                    ref_time_met = True
                    break
            done = ref_time_met or bb_commits.get('isLastPage', False) == True
        return commits[::-1], stop_sha

    def getRecentPullRequests(self, repo_url, ref_time):
        pr_resource = 'pullrequests' if self.cloud else 'pull-requests'
        pr_url = "%s/%s?state=ALL&pagelen=50" % (repo_url, pr_resource)
        try:
            result = self.getFromBitbucket(pr_url)
        except Exception as ex:
            ex.report()
            return []
        pull_requests = []
        for pr_info in result['values']:
            pull_request = BitbucketPullRequest(pr_info)
            if pull_request.creation_timestamp.epoch_seconds >= ref_time.epoch_seconds:
                pull_request.bb_helper = self
                pull_requests.append(pull_request)
        return pull_requests

    def getCommits(self, commits_url):
        try:
            result = self.getFromBitbucket(commits_url)
        except Exception as ex:
            ex.report()
            return []
        return result['values']


    # def getChange(self, bburl, sha):
    #     # https://developer.atlassian.com/static/rest/bitbucket-server/4.14.4/bitbucket-rest.html#idm46478324963760
    #     # the bburl must be of the format http://server:port/rest/api/1.0/projects/<projectkey>/repos/<repo_name>
    #     pending_operation = "Retrieving the Change for sha '%s'"
    #
    #     self.log.info(pending_operation % sha)
    #
    #     url = "%s/changes/?until=%s" %(bburl, sha)
    #
    #     self.log.debug("Using url: %s" % url)
    #
    #     response = self.getFromBitbucket(url)
    #
    #     if 'errors' in response:
    #         self.log.error(response['errors'][0]['message'])
    #     else:
    #         self.log.debug("  path         : %s" % response['values'][0]['path']['name'])
    #         self.log.debug("  contentId    : %s" % response['values'][0]['contentId'])
    #         self.log.debug("  fromContentId: %s" % response['values'][0]['fromContentId'])
    #
    #     return response
