# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

class VCSConnection(object):

    def __init__(self, logger):
        self.log      = logger
        self.config   = None
        self.username = None
        self.password = None
        self.username_required = True
        self.password_required = True
        self.build_selectors   = []
        self.log.info("Initializing %s connection version %s" % (self.name(), self.version()))

    def name(self):
        """
            abstract, provider should return a non-empty string with name of the specific connector
        """
        raise NotImplementedError("A VCSConnection subclass must implement the name method")

    #Placeholder to put the version of the connector
    def version(self):
        """
            abstract, provider should return a non-empty string with version 
            identification of the specific connector
        """
        raise NotImplementedError("All descendants of the VCSConnection class need to implement version()")
        #Should return a string

    def connect(self):
        """
            Returns True or False depending on whether a connection was "established".
            As many connectors are stateless, the "establishment of a connection" might
            just mean that the target and credentials are adequate to post a request and
            receive a non-error response.
        """
        raise NotImplementedError("All descendants of the VCSConnection class need to implement connect()")
        #Should return True or False

    def disconnect(self):
        """
            Returns True or False depending on whether an existing connection was disconnected
            successfully.
            As many connectors are stateless, the disconnection may be as easy as 
            resetting an instance variable to None
        """
        raise NotImplementedError("All descendants of the VCSConnection class need to implement disconnect()")
        #Should return True or False

    def fieldExists(self, field_name):
        """
            Return a boolean truth value (True/False) depending on whether the targeted
            field_name exists for the current connection on a Build
        """
        raise NotImplementedError("All descendants of the VCSConnection class need to implement fieldExists(field_name)")


    def validate(self):
        """
        """
        raise NotImplementedError("All descendants of the VCSConnection class need to implement a validation method")


    def internalizeConfig(self, config):
        """
            config has already been read, it comes to us here as
            a dict with information relevant to this connection
        """
        self.config = config
        self.username = config.get('Username', config.get('User', False))
        self.password = config.get('Password', False)
        self.proxy_protocol = config.get('ProxyProtocol', 'http')
        self.proxy_server   = config.get('ProxyServer',   False)
        self.proxy_port     = config.get('ProxyPort',     False)
        self.proxy_username = config.get('ProxyUsername', config.get('ProxyUser', False))
        self.proxy_password = config.get('ProxyPassword', False)

        self.debug    = config.get('Debug', False)
        #print("Connection debug value is {0}".format(self.debug))
        if self.debug:
            self.log.setLevel('DEBUG')


    def getRecentCommits(self, ref_time):
        """
            Finds items that have been created since a reference time (ref_time is in UTC) 
            and applying all specified BuildSelector conditions.

            Concrete subclasses must implement this method and return a list of qualified items.
        """
        problem = "All descendants of the VCSConnection class need to implement getRecentCommits(ref_time)"
        raise NotImplementedError(problem)
