# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import os
import re
import socket
import time
import copy

from vcsconn.connection    import VCSConnection
from vcsconn.ghql_adapter  import GitHubGraphQLAdapter
from vcsconn.graphql_query import GraphQL_Query
from vcsconn.github_changeset import GitHubChangeset
from vcsconn.github_pullrequest import GitHubPullRequest
from vcsconn.utils.eif_exception import ConfigurationError, OperationalError
from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT

############################################################################################

__version__ = "2.0.5"

ACTION_WORD_PATTERN    = re.compile(r'[A-Z][a-z]+')
ARTIFACT_IDENT_PATTERN = re.compile(r'(?P<art_prefix>[A-Z]{1,4})(?P<art_num>\d+)')
VALID_ARTIFACT_ABBREV = None  # set later after config values for artifact prefixes are known

############################################################################################

class GitHubConnection(VCSConnection):
    def __init__(self, config, logger):
        super().__init__(logger)
        self.gha = None
        self.internalizeConfig(config)
        self.backend_version = ""

    def name(self):
        return "GitHub"

    def version(self):
        global __version__
        return __version__

    def getBackendVersion(self):
        """
            Conform to Connection subclass protocol and provide the version of the system
            we are "connecting" to.
        """
        return self.backend_version

    def internalizeConfig(self, config):
        super().internalizeConfig(config)
        self.protocol     = config.get('Protocol', 'https')
        self.server       = config.get('Server', 'api.github.com')
        self.repo_server  = 'github.com' if self.server == 'api.github.com' else self.server
        self.port         = config.get('Port', 8080)
        self.prefix       = config.get('Prefix', '')
        self.gh_token     = config.get('Token', '')
        self.organization = config.get('Organization', None)
        self.repo_spec    = config.get('Repository',   None)
        self.debug        = config.get('Debug',        False)
        self.cert_file    = config.get('CertFile',     None)
        #self.max_commits  = config.get('MaxCommits',   1000)
        self.base_url     = '{0}://{1}:{2}/graphql'.format(self.protocol, self.server, self.port)
        self.repo_site = '{0}://{1}'.format(self.protocol, self.repo_server)
        if self.username:
            if self.gh_token:
                cred = self.gh_token
            else:
                cred = self.password
            self.creds = (self.username, cred)
            if not self.organization: self.organization = self.username  # use the username as the organization if Organization not specified in config
        else:
            self.creds = None

        #https_proxy = os.environ.get('https_proxy', None) or os.environ.get('HTTPS_PROXY', None)
        #if https_proxy not in ["", None]:
        #    self.log.info("Proxy for GitHub connection:  %s" % https_proxy)

        self.http_proxy = {}
        if self.proxy_server:
            proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
            self.http_proxy = {self.protocol : proxy}
            self.log.info("Proxy for GitHub connection:  %s" % proxy)

        valid_config_items = ['Server', 'Protocol', 'Port',
                              'Username', 'User', 'Password', 'Token',
                              'ProxyProtocol', 'ProxyServer', 'ProxyPort', 'ProxyUser', 'ProxyUsername', 'ProxyPassword',
                              'Organization', 'Repository', 'Debug'
                             ]

        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "GitHub section of the config contained these invalid entries: %s" % ", ".join(
                invalid_config_items)
            raise ConfigurationError(problem)

        if self.cert_file:
            os.environ['REQUESTS_CA_BUNDLE'] = self.cert_file  # '/etc/ssl/certs/cacert.pem' or other location...

    def connect(self):
        """
        """
        self.log.info("Connecting to GitHub")

        self.backend_version = self._getGitHubVersion()
        self.log.info("Connected to GitHub server: %s running at version %s" % (self.server, self.backend_version))
        self.log.info("Url: %s" % self.base_url)
        # we get a copy of self.config so we can set the Organization value to send to GraphQLAdapter which will internalize it
        # but we don't want to sully our config because it would then goof up the validation
        gha_config = copy.copy(self.config)
        gha_config['Organization'] = self.organization
        self.gha = GitHubGraphQLAdapter(gha_config, self.log)
        org_key = self.config.get('Organization', None)
        if org_key:
           self.obtainGitHubOrganizations()
        else:
            self.organizations = [self.organization]  # to prevent burn-out in validation
        return True

    def _getGitHubVersion(self):
        """
            There may not be a way to get the GitHub version as GH is a SaaS entity, and may not expose a "version"
            If so, fake it
        """
        version = '2017'
        return version

    def disconnect(self):
        """
            Just reset our jenkins instance variable to None
        """
        self.gha = None

    def validate(self):
        """
            Make sure any requisite conditions are satisfied.
            Are credentials needed and supplied?
        """
        satisfactory = True
        if self.username:
            self.log.debug('GitHubConnection - user entry "%s" detected in config' % (self.username))
        else:
            satisfactory = False
            self.log.error("No User value was provided in the GitHub section of the config")

        if self.gh_token:
            self.log.debug("GitHub Token will be used")
        elif self.password:
            self.log.debug("GitHub Password will be used")
        else:
            satisfactory = False
        # check to see if the self.organization (from the config) is present in the self.organizations list
        if self.organization:
            matches = [org for org in self.organizations if org.lower() == self.organization.lower()]
            if not matches:
                self.log.error("The Organization value provided '%s' does not exist or is not accessible for the credentials provided" % self.organization)
                return False
            else:
                self.organization = matches[0]

        # now go get the repositories that exist under the specified Organization
        # and then pass them through the filters of Include and Exclude to arrive at a list
        # of actionable repository names
        accessible_repos = self.obtainGitHubAccessibleRepositories()
        if not self.repo_spec or len(self.repo_spec) == 0:
            processable_repos  = accessible_repos
        else:
            processable_repos  = accessible_repos
            if 'Include' in self.repo_spec and self.repo_spec['Include']:
                processable_repos = self.qualifyRepos(accessible_repos,   self.repo_spec['Include'], True)

            if 'Exclude' in self.repo_spec and self.repo_spec['Exclude']:
                processable_repos = self.qualifyRepos(processable_repos,  self.repo_spec['Exclude'], False)

        if not processable_repos:
            self.log.error("All repos specified in your config are not found among the accessible repos")
            return False
        self.processable_repos = processable_repos

        return satisfactory


    def qualifyRepos(self, repo_names, qualifications, include):
        if not qualifications:
            return repo_names

        qualification_items = re.split(r',\s*', qualifications)
        amended_repo_names = []
        if not include:
            amended_repo_names = repo_names[:]
        for repo_name in repo_names:
            for qual in qualification_items:
                # if qual has no '*' char do a simple string compare and use repo that logic result
                # otherwise change the shell wildcard char '*' into a regex wild card '.*' and use the regex result
                if '*' not in qual:
                    if repo_name.upper() == qual.upper():
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
                else:
                    qual_regex = '^%s' % qual.replace('*', '.*')
                    if re.search(qual_regex, repo_name, re.I): # re.I makes repo name case insensitive
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break

        return amended_repo_names


    def extractNestedStruct(self, root, key_chain, node_name, attribute):
        struct = root
        for key in key_chain:
            if struct.get(key, None):
                struct = struct[key]
            else:
                return []
        items = [item[node_name].get(attribute) for item in struct]
        return items


    def obtainGitHubOrganizations(self):
        """
             Utilize the GitHub GraphQL endpoint to obtain all visible/accessible GitHub Organizations
        """
        orgs_query = GraphQL_Query.constructOrganizationsQuery()
        try:
            result = self.gha.makeQueryRequest(orgs_query)
        except Exception as exc:
            raise ConfigurationError(exc)
        key_chain = ['data', 'viewer', 'organizations', 'edges']
        org_paths = self.extractNestedStruct(result, key_chain, 'org', 'resourcePath')
        short_names = [path[1:] for path in org_paths]
        self.organizations = short_names

    # def getUserOrOrganization(self):
    #     return self.organization if self.organization else self.username

    def obtainGitHubAccessibleRepositories(self):
        #user_org = self.getUserOrOrganization()
        repos_query = GraphQL_Query.constructOrganizationRepositoriesQuery(self.organization)
        try:
            repos = self.gha.getAllQueryPages(repos_query)
        except Exception as exc:
            raise ConfigurationError(exc.args[0])
        repo_names = sorted([repo['name'] for repo in repos])
        return repo_names

    def getRecentCommits(self, ref_time):
        #user_org = self.getUserOrOrganization()
        commits = self.getRecentRepositoryActivity(ref_time)
        return commits


    def getRecentPullRequests(self, ref_time):
        """
            Select only those PullRequests whose CreatedAt date is >= ref_time_iso.
        """
        def qualified(pr, ref_time):
            if not pr['commits'] or not pr['commits']['nodes']:
                return False
            return TimeStamp.fromFormattedString(pr['createdAt'], format=ISO_FORMAT).epoch_seconds >= ref_time.epoch_seconds

        for attempt in [1,2,3]:
            pull_requests = [] #raw return from graphql
            try:
                raw = self.gha.getPullRequestPages(self.processable_repos, ref_time)

            except Exception as e:
                if attempt <= 2:
                    time.sleep(attempt * 5)
                    self.log.debug("Attempt # %d to getRecentPullRequests from GitHub GraphQL" % (attempt+1))
                    continue
                else:
                    break

            filtered_prs = [GitHubPullRequest(pr) for all_prs in raw.values() for pr in all_prs if qualified(pr, ref_time)]
            return filtered_prs


    def getRecentRepositoryActivity(self, ref_time):
        """
            Obtain all Repository commits that have happened in the repositories owned by organization  at or after the ref_time parameter
            which is a struct_time object of:
               (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst)
            Return a dict keyed by repo_name, associated values will be a list of GitHubChangeset instances
        """
        repo_commits = {}
        pending_operation = "Detecting recently added GitHub commits in the %s organization (on or after %s)"
        self.log.info(pending_operation % (self.organization, ref_time.asLogStamp()))

        for attempt in [1,2,3]:
            active_repositories = [] #raw return from graphql
            try:
                active_repositories = self.gha.getRepoAndCommitPages(self.processable_repos, ref_time, branch='master')

            except Exception as e:
                if attempt <= 2:
                    time.sleep(attempt * 5)
                    self.log.debug("Attempt # %d to getRecentRepositoryActivity from GitHub GraphQL" % (attempt+1))
                    continue
                else:
                    break

            total_commits = 0
            for repo, commits in active_repositories.items():
                repo_commits[repo] = sorted([GitHubChangeset(info) for info in commits])
                [info.completePath(self.repo_site) for info in repo_commits[repo]]
                total_commits += len(repo_commits[repo])
            self.log.info('Recent GitHub commits (since %s) detected: %d' % (ref_time.asLogStamp(), total_commits))
            return repo_commits

        problem = "Error in getRecentRepositoryActivity; GitHub GraphQL response potentially caused by reference_time too far in the past.  "
        remedy  = "Please try again later or consider changing the time file value to a value more recent than  %s" % ref_time.asLogStamp()
        raise OperationalError("Unable to retrieve GitHub commit info. %s" % (problem + remedy))

