# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import os
import re

import requests

from vcsconn.connection          import VCSConnection
from vcsconn.utils.eif_exception import ConfigurationError, OperationalError
from vcsconn.tfsgit_helper       import TFSGitHelper

############################################################################################

class TFSGitConnection(VCSConnection):

    TFSGIT_CONN_VERSION = '1.0.0'

    def __init__(self, config, logger):
        super().__init__(logger)
        self.server         = None

        self.internalizeConfig(config)
        self.backend_version          = None
        self.processable_repos        = []
        self.repos_by_name            = {}
        self.tfsgit_helper            = None
        self.api_version = {'ProjectCollection' : '1.0-preview.2',
                            'TeamProject'       : '1.0',
                            'Repositories'      : '1.0',
                            'Commits'           : '1.0',
                            'PullRequests'      : '3.0'
                           }
        self.base_url = "%s://%s:%s" % (self.protocol, self.server, self.port)
        if self.instance:
            self.base_url += "/%s" % self.instance
        return

    def name(self):
        return "TFSGit"

    def version(self):
        return self.TFSGIT_CONN_VERSION

    def internalizeConfig(self, config):
        super().internalizeConfig(config)
        self.protocol         = config.get('Protocol',  'http')
        self.server           = config.get('Server',     None)
        self.port             = config.get('Port',      '80' if self.protocol == 'http' else '443')
        self.instance         = config.get('Instance',   None)
        #self.username         = config.get('Username',   None)  # taken care of in Connection superclass method
        self.token            = config.get('Token',      None)   # This would be a TFS 2017 personal security token
        self.cert_file        = config.get('CertFile',   None)
        self.collection       = config.get('Collection', 'DefaultCollection')
        self.project          = config.get('TeamProject', None)
        self.max_items        = config.get('MaxItems',   100)
        self.backend_version  = None
        self.api_version      = config.get('APIversion', 1.0)  # config.get('APIversion', self.backend_version)
        self.debug            = config.get('Debug',      False)
        self.repo_spec        = config.get('Repository', None)

        if self.repo_spec:
            if not 'Include' in self.repo_spec or not self.repo_spec['Include'] or not self.repo_spec['Include'].strip():
                self.repo_spec['Include'] = None
            if not 'Exclude' in self.repo_spec or not self.repo_spec['Exclude'] or not self.repo_spec['Exclude'].strip():
                self.repo_spec['Exclude'] = None
            if not self.repo_spec['Include'] and not self.repo_spec['Exclude']:
                self.repo_spec = None

        self.http_proxy = {}
        if self.proxy_server:
            proxy = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                proxy = "%s://%s:%s@%s:%s" % (
                    self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
            self.http_proxy = {self.protocol: proxy}
            self.log.info("Proxy for TFSGit connection:  %s" % proxy)

        # Fixup things we need
        if self.server.__class__.__name__ == 'str':
            self.server = re.sub(r'^https?:\/', '', self.server)

        if self.debug and self.debug.lower() in ['false', 'no', 'off', '0']:
            self.debug = False
        if self.debug and self.debug.lower() in ['true', 'yes', 'on', '1']:
            self.debug = True

        if self.username and self.token:
            self.auth = (self.username, self.token)

        if self.cert_file:
            os.environ['REQUESTS_CA_BUNDLE'] = self.cert_file  # '/etc/ssl/certs/cacert.pem' or other location...

    def connect(self):
        self.log.info("Connecting to TFSGit...")
        uri = self.getTFSGitCollectionURL(self.collection)
        try:
            headers = {'Content-Type': 'application/json'}
            response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
            result = response.json()
        except Exception as e:
            problem = "Unable to connect to TFS Git at %s \n        Test this url in your browser. If it does not work there, it will not work here" % (uri)
            raise ConfigurationError(problem)
        if 'name' not in result:
            error_text = 'Collection %s not found' % self.config['Collection']
            if 'message' in result:
                error_text = result['message']
            raise ConfigurationError(error_text)
        self.log.info("Connected to TFS Git at %s" % self.server)
        self.connected = True
        return True

    def disconnect(self):
        self.connected = False


    def validate(self):

        satisfactory = True

        if not self.server:
            self.log.error("'Server:' is required in the TFSGit section of the config file")
            return False

        if not self.username:
            self.log.error("'Username:' is required in the TFSGit section of the config file")
            return False

        if not self.token:
            self.log.error("'Token:' is required in the TFSGit section of the config file")
            return False

        self.tfsgit_helper = TFSGitHelper(self.username, self.token, self.log, self.http_proxy)

        project = self.getTFSGitProject(self.collection, self.project)
        if not project:
            self.log.error("'TeamProject:' is required in the TFSGit section of the config file")
            return False

        self.log.info("TFSGit Connection details:  %s://%s:%s/%s/%s" % (self.protocol, self.server, self.port, self.collection, self.project))

        accessible_repos = self.obtainTFSGitAccessibleRepositories(self.collection, self.project)

        if not self.repo_spec or len(self.repo_spec) == 0:
            processable_repos = accessible_repos
        else:
            processable_repos = accessible_repos
            if 'Include' in self.repo_spec and self.repo_spec['Include']:
                processable_repos = self.qualifyRepos(accessible_repos, self.repo_spec['Include'], True)
                if not processable_repos:
                    self.log.error("Repositories specified in your config are not found among the accessible repos: %s" % ", ".join(accessible_repos))
                    return False

            if 'Exclude' in self.repo_spec and self.repo_spec['Exclude']:
                prior_repos = processable_repos[:]
                processable_repos = self.qualifyRepos(prior_repos, self.repo_spec['Exclude'], False)
                excluded_repos = [repo for repo in prior_repos if repo not in processable_repos]
                self.log.debug("These repos were excluded based on your Repositories -> Exclude criteria: %s" % ", ".join(excluded_repos))

        self.processable_repos = processable_repos
        if not processable_repos:
            self.log.error("There are no repositories that meet your criteria")
            satisfactory = False
        else:
            self.log.debug("These repositories will be processed: %s" % ", ".join(processable_repos))
        return satisfactory

    def getTFSGitCollectionURL(self, collection):
        url = "%s/_apis/projectcollections/%s?api-version=%s" % (self.base_url, collection, self.api_version['ProjectCollection'])
        return url

    def getTFSGitProjectURL(self, collection, project):
        url = "%s/%s/_apis/projects/%s?api-version=%s" % (self.base_url, collection, project, self.api_version['TeamProject'])
        return url

    def getTFSGitRepositoriesURL(self, collection, project):
        url = "%s/%s/%s/_apis/git/repositories?api-version=%s" % (self.base_url, collection, project, self.api_version['Repositories'])
        return url

    def getTFSGitRepositoryCommitsURL(self, collection, project, repository, ref_time):
        url = "%s/%s/%s/_apis/git/repositories/%s/commits?api-version=%s&fromDate=%s" % \
               (self.base_url, collection, project, repository, self.api_version['Commits'], ref_time.asISOString())
        return url

    def getTFSGitRepositoryPullRequestsURL(self, collection, project, repository):
        url = "%s/%s/%s/_apis/git/repositories/%s/pullRequests?api-version=%s&status=all" % \
               (self.base_url, collection, project, repository, self.api_version['PullRequests'])
        return url

    def getTFSGitProject(self, collection, project):
        project_url = self.getTFSGitProjectURL(collection, project)
        result = self.tfsgit_helper.getFromTFSGit(project_url)
        if 'name' not in result:
            return None
        return result['name']

    def obtainTFSGitAccessibleRepositories(self, collection, project):
        repos_url = self.getTFSGitRepositoriesURL(collection, project)
        result = self.tfsgit_helper.getFromTFSGit(repos_url)
        accessible_repos = [repo['name'] for repo in result['value'] for key in repo.keys() if key == 'name']
        self.repos_by_name = {}

        for repo in result['value']:
            self.repos_by_name[repo['name']] = repo
        return accessible_repos

    def qualifyRepos(self, repo_names, qualifications, include):
        qualification_items = re.split(r',\s*', qualifications)
        amended_repo_names = []
        if not include:
            amended_repo_names = repo_names[:]
        for repo_name in repo_names:
            for qual in qualification_items:
                # if qual has no '*' char do a simple string compare and use repo that logic result
                # otherwise change the shell wildcard char '*' into a regex wild card '.*' and use the regex result
                if '*' not in qual:
                    if repo_name.upper() == qual.upper():
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
                else:
                    qual_regex = '^%s' % qual.replace('*', '.*')
                    if re.search(qual_regex, repo_name, re.I): # re.I makes repo name case insensitive
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break

        return amended_repo_names


    def getRecentCommits(self, ref_time):
        """
            Obtain all Changesets created in TFS at or after the ref_time parameter which is a TimeStamp instance
        """
        all_commits = {}  # will be keyed by repo name

        pending_operation = "Detecting recently added TFS Git commits (added on or after %s)"
        self.log.info(pending_operation % ref_time.asLogStamp())
        for repo in self.processable_repos:
            all_commits[repo] = self.getRecentRepositoryCommits(repo, ref_time)
        return all_commits

    def getRecentRepositoryCommits(self, repo, ref_time):
        commits_url = self.getTFSGitRepositoryCommitsURL(self.collection, self.project, repo, ref_time)
        changesets = self.tfsgit_helper.getQualifiedCommits(commits_url, repo)
        chrono_ordered_changesets = sorted(changesets, key=lambda x: x.commit_timestamp)
        return chrono_ordered_changesets

    def getRecentPullRequests(self, ref_time):
        """
            Select only those PullRequests whose creationDate is >= ref_time_iso.
            iterate over the self.processable_repos to arrive at a single list of Pull Requests.
        """
        processable_prs = []
        for repo in self.processable_repos:
            pullrequests_url   = self.getTFSGitRepositoryPullRequestsURL(self.collection, self.project, repo)
            pull_requests      = self.tfsgit_helper.getQualifiedPullRequests(pullrequests_url, ref_time)
            chrono_ordered_prs = sorted(pull_requests, key=lambda x: x.creation_timestamp)
            processable_prs.extend(chrono_ordered_prs)

        return processable_prs
