# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import os
import re

import requests
from vcsconn.connection import VCSConnection
from vcsconn.utils.eif_exception import ConfigurationError, OperationalError

from vcsconn.vsts_helper import VSTSHelper
from vcsconn.vsts_pullrequest import  VSTSPullRequest
from vcsconn.vsts_changeset   import  VSTSChangeset
from vcsconn.vsts_repository  import  VSTSRepository
from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT, READABLE_FORMAT

############################################################################################

class VSTSConnection(VCSConnection):

    VSTS_CONN_VERSION = '2.0.5'

    def __init__(self, config, logger):
        super().__init__(logger)
        self.server         = None
        self.projectkey     = None
        self.internalizeConfig(config)
        self.backend_version     = None
        self.processable_repos   = []
        self.vsts_helper         = None
        return

    def name(self):
        return "VSTS"

    def version(self):
        return self.VSTS_CONN_VERSION


    def connect(self):
        self.log.info("Connecting to VSTS...")
        self.backend_version = None
        problem = ""

        if not self.config.get('TeamProject', False):
            problem = 'TeamProject is not specified in  VSTS section of the configuration file. Only one of them can be present.'
            raise ConfigurationError(problem)
        self.base_url = "%s://%s/%s" % (self.protocol, self.server, self.collection)
        #if self.config.get('TeamProject', False):
        #your-instance.visualstudio.com/DefaultCollection/_apis/projects?api-version=3.0
        uri = "%s/_apis/projects?api-version=%s" % (self.base_url, self.api)
        try:
            headers = {'Content-Type': 'application/json'}
            response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
            result = response.json()
            projects = result['value']
            if not [project['name'] for project in projects if project['name'] == self.project]:
                problem = "TeamProject provided in the VSTS section of the configuration file is invalid."
                raise ConfigurationError(problem)
            else:
                self.backend_version = self.api
                self.log.info("Connected to VSTS at %s" % self.server)
        except Exception as e:
            problem = "Unable to connect to VSTS at %s. %s" % (uri, e)
            raise ConfigurationError(problem)

        if not self.backend_version:
            self.log.error(problem)
            return False
        return self.backend_version


    def internalizeConfig(self, config):
        super().internalizeConfig(config)
        self._setConnectionDefaults(config)

        self.protocol         = config.get('Protocol', 'https')
        self.server           = config.get('Server',    None)
        #self.port             = config.get('Port',      443)  # we assume all VSTS interaction is going to be https
        self.custom_prefix    = config.get('Prefix',    None)
        #self.username         = config.get('Username',  None) # taken care of in Connection superclass method
        self.token            = config.get('Token',  None)
        self.cert_file        = config.get('CertFile',  None)
        self.collection       = config.get('Collection','DefaultCollection')
        self.project          = config.get('TeamProject', "")

        self.backend_version  = None
        self.debug            = config.get('Debug'     , False)
        self.repo_spec        = config.get('Repository', None)

        self.http_proxy = {}
        if self.proxy_server:
            proxy = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                proxy = "%s://%s:%s@%s:%s" % (
                    self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
            self.http_proxy = {self.protocol: proxy}
            self.log.info("Proxy for VSTS connection:  %s" % proxy)

        # Fixup things we need
        if self.server.__class__.__name__ == 'str':
            self.server = re.sub(r'^https?:\/', '', self.server)

        if self.debug and self.debug.lower() in ['false', 'no', 'off', '0']:
            self.debug = False
        if self.debug and self.debug.lower() in ['true', 'yes', 'on', '1']:
            self.debug = True

        self.project_required = True

        if self.username and self.token:
            self.auth = (self.username, self.token)

        if self.cert_file:
            os.environ['REQUESTS_CA_BUNDLE'] = self.cert_file  # '/etc/ssl/certs/cacert.pem' or other location...

    def _setConnectionDefaults(self, config):
        if config.get('Server', None) and config.get('Server', None).endswith('.visualstudio.com'):
            self.vcs   = 'git'
            self.api   = '3.0'
            config['Protocol'] = 'https'


    def disconnect(self):
        pass

    def validate(self):
        satisfactory = True
        self.vsts_helper = VSTSHelper(self.username, self.token, self.log, self.http_proxy)
        accessible_repos = []
        accessible_repo_names   = []
        processable_repo_names = []

        if not self.project:
            self.log.error("'TeamProject' is required in the VSTS section of the config file")
            return False
        repos_url = "%s/%s/_apis/%s/repositories?api-version=%s" % (self.base_url, self.project, self.vcs, self.api)
        accessible_repos = self.obtainVSTSAccessibleRepositories(repos_url)
        accessible_repo_names = [repo['name'] for repo in accessible_repos for key in repo.keys() if key == 'name']

        if not self.repo_spec or len(self.repo_spec) == 0:
            processable_repo_names = accessible_repo_names
        else:
            processable_repo_names = accessible_repo_names
            if 'Include' in self.repo_spec:
                processable_repo_names = self.qualifyRepos(accessible_repo_names, self.repo_spec['Include'], True)
                if not processable_repo_names:
                    self.log.error("All repos specified in your config are not found among the accessible repos")
                    return False

            if 'Exclude' in self.repo_spec:
                prior_repos = processable_repo_names[:]
                processable_repo_names = self.qualifyRepos(prior_repos, self.repo_spec['Exclude'], False)
                excluded_repo_names = [repo for repo in prior_repos if repo not in processable_repo_names]
                self.log.debug("These repos were excluded based on your Repositories -> Exclude criteria: %s" % ", ".join(excluded_repo_names))

        target_repos = [repo for repo in accessible_repos if repo['name'] in processable_repo_names]
        # The repo objects below from target_repos do not have '_links' property, hence they have to be inflated
        repo_info = []
        for repo in target_repos:
            try:
                vsts_repo_info = self.vsts_helper.getFromVSTS(repo['url'])
                repo_info.append(vsts_repo_info)
            except:
                self.log.error("Unable to obtain complete information on VSTS repo: %s" % repo['name'])

        self.processable_repos = []
        for vsts_repo_item in repo_info:
            ri_keys = list(vsts_repo_item.keys())
            missing_keys = [key for key in ['id', 'name', 'project', '_links'] if key not in ri_keys]
            if missing_keys:
                self.log.error("VSTS JSON data for repo item %s is missing these data keys: %s" % (repr(vsts_repo_item), ", ".join(missing_keys)))
                continue
            try:
                vsts_repo = VSTSRepository(vsts_repo_item)
                self.processable_repos.append(vsts_repo)
            except Exception as ex:
                actual_error = ex.args[0]
                self.log.error("Unable to instantiate a VSTSRepository for %s, %s" % (vsts_repo_item['name'], actual_error))

        if not self.processable_repos:
            self.log.error("There are no repositories that meet your criteria")
            satisfactory = False
        else:
            self.log.debug("These repositories will be processed: %s" % ", ".join(processable_repo_names))
        return satisfactory

    def obtainVSTSAccessibleRepositories(self, target_url):
        self.log.info("  VSTS connect attempt using uri=%s" % target_url)
        accessible_repos = self.vsts_helper.getAllRepos(target_url)
        return accessible_repos

    def qualifyRepos(self, repo_names, qualifications, include):

        qualification_items = re.split(r',\s*', qualifications)
        amended_repo_names = []
        if not include:
            amended_repo_names = repo_names[:]
        for repo_name in repo_names:

            match = False
            for qual in qualification_items:
                # if qual has no '*' char do a simple string compare and use repo that logic result
                # otherwise change the shell wildcard char '*' into a regex wild card '.*' and use the regex result
                if '*' not in qual:
                    if repo_name.upper() == qual.upper():
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
                else:
                    qual_regex = '^%s' % qual.replace('*', '.*')
                    if re.search(qual_regex, repo_name, re.I): # re.I makes repo name case insensitive
                        if include:
                            amended_repo_names.append(repo_name)
                        else:
                            amended_repo_names.remove(repo_name)
                        break
        return amended_repo_names


    def getRecentCommits(self, ref_time, last_commit_sha=None):
        """
            Obtain all Changesets created in VSTS at or after the ref_time parameter which is a Time instance
        """
        all_commits = {}  # will be keyed by repo name

        pending_operation = "Detecting recently added VSTS Commits (added on or after %s)"
        self.log.info(pending_operation % ref_time.asLogStamp())
        for repo in self.processable_repos:
            all_commits[repo.name] = self.getRecentRepositoryCommits(repo, ref_time)
        return all_commits

    def getRecentRepositoryCommits(self, repo, ref_time):
        url = "%s?api-version=%s" %  (repo.commits_url, self.api)
        changesets = self.vsts_helper.getQualifiedCommits(url, ref_time)
        chrono_ordered_changesets = sorted(changesets, key=lambda x: x.commit_timestamp)
        return chrono_ordered_changesets

    def getRecentPullRequests(self, ref_time):
        """
            Select only those PullRequests whose CreatedAt date is >= ref_time_iso.
            iterate over the self.processable_repos to arrive at a single list of Pull Requests.
            We don't care/mind that the PullRequests are not organized into a struct keyed by repo_slug.
        """
        processable_prs = []
        for repo in self.processable_repos:
            url = "%s?api-version=%s" % (repo.pullrequests_url, self.api)
            pull_requests = self.vsts_helper.getQualifiedPullRequests(url, ref_time)
            chrono_ordered_prs = sorted(pull_requests, key=lambda x: x.creation_timestamp)
            processable_prs.extend(chrono_ordered_prs)

        return processable_prs
